
--
--  Gossipmonger
--    by Tuhljin
--

local THIS_VERSION = GetAddOnMetadata("Gossipmonger", "Version");
local THIS_TITLE = GetAddOnMetadata("Gossipmonger", "Title")

local L = GOSSIPMONGER_STRINGS

local CYAN = "|cff00ffff"
local YELLOW = "|cffffff00"
local YELLOW_LT = "|cffffff99"
local R, G, B = 0.494, 0.729, 1
local SCTColors

local minimumGossipHeight, maximumGossipHeight = 150, 414
local incGossipHeight = 119

local ChatIcon = "|TInterface\\GossipFrame\\GossipGossipIcon:0:0:2:-2|t"

Gossipmonger = {};

Gossipmonger.DefaultSettings = {
  Version = THIS_VERSION,
  skipIsDefault = true,
  informOnSkip_SCT = false,
  informOnSkip_log = false,
  alwaysInformWithName = false,
  showGossipMirrorOnSkip = true,
  delayShowingMirror = true,
  showGossipMirrorOnQuestSelect = true,
  greetingModifierKey = 2,
  Filter_Standard = true,
};

local CurrentNPC, CurrentNPC_gossip
local SafeMirrorOptionsClick
local GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip  -- For the GossipFrame_OnEvent hook
local IgnoreQuestFinished, FollowupFunction_filtername

local function chatprint(msg)
  DEFAULT_CHAT_FRAME:AddMessage(msg, R, G, B);
end

local TQD, preTQD

local function prepGossipOptionDisplay(id, baseButton, selected)
  local name = "Gossipmonger_OptionDisplay"..id
  local titleButton = _G[name]
  if (not titleButton) then
    titleButton = CreateFrame("Button", name, Gossipmonger_MirrorFrame_GreetingScrollFrameScrollChild, "Gossipmonger_OptionDisplay")
    if (id == 1) then
      titleButton:SetPoint("TOPLEFT", Gossipmonger_MirrorFrame_GreetingText, "BOTTOMLEFT", -10, -20)
    else
      local prevButton = _G["Gossipmonger_OptionDisplay"..(id - 1)]
      titleButton:SetPoint("TOPLEFT", prevButton, "BOTTOMLEFT", 0, -3)
    end
  end

  titleButton.baseButton = baseButton
  local baseIcon = _G[baseButton.."GossipIcon"]
  baseButton = _G[baseButton]

  if (baseButton:IsShown()) then
    titleButton:Show()

    if (TRIVIAL_QUEST_DISPLAY ~= preTQD) then
      TQD = TRIVIAL_QUEST_DISPLAY:gsub("%(","%%(")
      TQD = TQD:gsub("%)","%%)")
      TQD = ("^"..TQD.."$"):gsub("%%s", "(.+)")
      preTQD = TRIVIAL_QUEST_DISPLAY
    end
    local NQD = ("^"..NORMAL_QUEST_DISPLAY.."$"):gsub("%%s", "(.+)")
    local text = baseButton:GetText()
    local _, baseTitle, taggedTitle
    _, _, baseTitle = text:find(TQD)
    _, _, taggedTitle = text:find(NQD)
    if (baseTitle) then  -- Low level quest
      titleButton.isLowLevel = true
    elseif (taggedTitle) then  -- Normal quest
      titleButton.isLowLevel = nil
      baseTitle = taggedTitle
    else
      baseTitle, taggedTitle = text, text
    end
    titleButton.title = baseTitle

    local title = selected and YELLOW .. taggedTitle .. "   " .. YELLOW_LT .. L.SELECTEDOPTION or
                  YELLOW .. taggedTitle
    titleButton:SetText(title);
    GossipResize(titleButton);   -- Calls a Blizzard function that currently simply does this:  titleButton:SetHeight( titleButton:GetTextHeight() + 2);
    titleButton:SetID( baseButton:GetID() )
    titleButton.type = baseButton.type  -- Set to "Available", "Active", or "Gossip"
    local icon = _G[name.."GossipIcon"]
    icon:SetTexture( baseIcon:GetTexture() )
    icon:SetVertexColor( baseIcon:GetVertexColor() )
    titleButton.selected = selected
    return titleButton:GetHeight(), true
  else
    titleButton.selected = nil
    titleButton:Hide()
  end
end

local function showGossipMirror(delay, name, text, selected, gossipType, noUpdate, watchQuest, numQuests)
  SafeMirrorOptionsClick = nil
  Gossipmonger_MirrorFrame_GreetingText:SetText(text);
  if (not noUpdate) then  -- noUpdate should be true when ShowGossipFrame already handled this
    GossipFrameUpdate() -- Update standard buttons; we'll grab the data from them rather than reinventing the
                        -- wheel. (It's hopefully more forward compatible this way, as well.)
  end
  -- Show used buttons:
  local optionsHeight, mirrorbtnIndex, h = -3, 1
  for i = 1, (GossipFrame.buttonIndex - 1) do  -- The actual number of buttons displayed is one less than given.
    h = prepGossipOptionDisplay(mirrorbtnIndex, "GossipTitleButton"..i, selected == mirrorbtnIndex)
    if (h) then
      optionsHeight = optionsHeight + h + 3
      mirrorbtnIndex = mirrorbtnIndex + 1
    end
  end
  
  Gossipmonger_MirrorFrame.selected = selected
  Gossipmonger_MirrorFrame.numOptions = mirrorbtnIndex - 1
  Gossipmonger_MirrorFrame.numQuests = numQuests or GetNumGossipAvailableQuests() + GetNumGossipActiveQuests()

  -- Hide the rest:
  local titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  while (titleButton) do
    titleButton:Hide()
    mirrorbtnIndex = mirrorbtnIndex + 1
    titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  end

  h = Gossipmonger_MirrorFrame_GreetingText:GetHeight() or 0
  local h2 = h + optionsHeight + incGossipHeight
  if (h2 < minimumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(minimumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  elseif (h <= Gossipmonger_MirrorFrame_GreetingScrollFrame:GetHeight() and h2 <= maximumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(h2)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  else
    Gossipmonger_MirrorFrame:SetHeight(maximumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Show()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel()
  end

  CurrentNPC, CurrentNPC_gossip = name, gossipType

  Gossipmonger_MirrorFrame_NpcNameText:SetText( CurrentNPC or CYAN..THIS_TITLE )
  if ( UnitExists("npc") ) then
    Gossipmonger_MirrorFrame_NpcNameText:SetTextColor(GameTooltip_UnitColor("npc"))
    SetPortraitTexture(Gossipmonger_MirrorFrame_Portrait, "npc");
    Gossipmonger_MirrorFrame.NPCName = CurrentNPC
  else
    Gossipmonger_MirrorFrame_Portrait:SetTexture("Interface\\QuestFrame\\UI-QuestLog-BookIcon");
    Gossipmonger_MirrorFrame.NPCName = nil
  end

  Gossipmonger_MirrorFrame_GreetingScrollFrame:SetVerticalScroll(0);
  if (not delay) then
    ShowUIPanel(Gossipmonger_MirrorFrame)
  end
  
  if (watchQuest) then
    if (watchQuest == true or Gossipmonger_Settings.showGossipMirrorOnQuestSelect) then
    -- watchQuest is set to 1 if the mirror should only automatically pops up if the settings tell it to do so,
    -- or true if it should always pop up (as is the case when a filter selects it). Either way, if the player
    -- clicked the vendor option, for example, instead of a quest, the mirror shouldn't pop up since it wasn't
    -- due to a filter selecting it.
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_DETAIL")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_PROGRESS")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_COMPLETE")
    end
  else
    Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_BINDER")
  end
end

local function BreakMirror()
-- We're not just hiding the mirror. We're "breaking" it away from the current NPC. (There's a distinction because
-- there may be times we want to have the option of showing it again w/o the disassociation that goes on here.)
  CurrentNPC, CurrentNPC_gossip, SafeMirrorOptionsClick = nil, nil, nil
  if (Gossipmonger_MirrorFrame:IsShown()) then
    HideUIPanel(Gossipmonger_MirrorFrame)
  end
  Gossipmonger_MirrorFrame:UnregisterAllEvents();
end

local function ShowMirror()
  if (not Gossipmonger_MirrorFrame:IsShown()) then
    ShowUIPanel(Gossipmonger_MirrorFrame)
  end
  SafeMirrorOptionsClick = true

  Gossipmonger_MirrorFrame:UnregisterAllEvents();
  Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("QUEST_FINISHED")
end

local function tooltip_StandardFilter(self, tooltip)
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP, nil, nil, nil, 1);
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP2, nil, nil, nil, 1);
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP3, nil, nil, nil, 1);
  tooltip:AddTexture("Interface\\GossipFrame\\GossipGossipIcon");
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP4, nil, nil, nil, 1);
end

local function UpdateLabel()
  if (Gossipmonger_Settings.skipIsDefault) then
    Gossipmonger_GreetingModDropdown:SetLabel(L.MODIFIERKEY_LABEL_DONOT)
  else
    Gossipmonger_GreetingModDropdown:SetLabel(L.MODIFIERKEY_LABEL_DO)
  end
end

local function UpdateStandardFilters()
  if (Gossipmonger_Settings.Filter_Standard) then
    for i,func in ipairs(Gossipmonger.StandardFilters) do
      Gossipmonger.EnableFilter(func)
    end
  else
    for i,func in ipairs(Gossipmonger.StandardFilters) do
      Gossipmonger.DisableFilter(func)
    end
  end
end

local function Init()
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Gossipmonger.OnEvent);

  local KEY, KEYTIP = L.KEY, L.MODIFIERKEY_TOOLTIP_KEY
  local KeySelectDropDown = {
    {
      text = KEY:format(L.ALT),
      tooltipTitle = KEY:format(L.ALT),
      tooltipText = KEYTIP:format(L.ALT),
      value = 0
    },
    {
      text = KEY:format(L.CTRL),
      tooltipTitle = KEY:format(L.CTRL),
      tooltipText = KEYTIP:format(L.CTRL),
      value = 1
    },
    {
      text = KEY:format(L.SHIFT),
      tooltipTitle = KEY:format(L.SHIFT),
      tooltipText = KEYTIP:format(L.SHIFT),
      value = 2
    },
    {
      text = L.MODIFIERKEY_NONE,
      tooltipTitle = L.MODIFIERKEY_NONE,
      tooltipText = L.MODIFIERKEY_NONE_TOOLTIP,
      value = 3
    };
  };

  local scttext = L.INFORMONSKIP_SCT
  if (not SCT) then  scttext = L.INFORMONSKIP_SCT_DISABLED:format(scttext);  end

  local panel, oldver = TjOptions.CreatePanel(THIS_TITLE, nil, {
  	title = THIS_TITLE.." v"..THIS_VERSION,
  	items = {
  		 { variable = "skipIsDefault", text = L.SKIPISDEFAULT, OnChange = UpdateLabel },
  		 { variable = "greetingModifierKey", type = "dropdown", name = "Gossipmonger_GreetingModDropdown",
  		   text = L.MODIFIERKEY_LABEL_DONOT, menu = KeySelectDropDown, width = 85 },
  		 { variable = "showGossipMirrorOnSkip", text = L.SHOWBUBBLE },
  		 { variable = "delayShowingMirror", text = L.SHOWBUBBLE_DELAY,
  		   tooltip = L.SHOWBUBBLE_DELAY_TOOLTIP, xOffset = 14 },
  		 { variable = "showGossipMirrorOnQuestSelect", text = L.SHOWBUBBLE_QUESTSELECT,
  		   tooltip = L.SHOWBUBBLE_QUESTSELECT_TOOLTIP, xOffset = 0, OnChange = BreakMirror },

  		 { type = "label", text = L.FILTEROPT_STANDARD_LABEL },
  		 { variable = "Filter_Standard", text = L.FILTEROPT_STANDARD, tooltip = tooltip_StandardFilter,
  		   xOffset = 14, OnChange = UpdateStandardFilters },

  		 { type = "label", text = L.INFORMONSKIP, xOffset = 0 },
  		 { variable = "informOnSkip_SCT", type = "checkboxsmall", text = scttext, xOffset = 14 },
  		 { variable = "informOnSkip_log", type = "checkboxsmall", text = L.INFORMONSKIP_CHAT },
  		 { variable = "alwaysInformWithName", type = "checkboxsmall", text = L.INFORMONSKIP_NAMEONLY,
  		   tooltip = L.INFORMONSKIP_NAMEONLY_TOOLTIP },
  	},
  	variables = "Gossipmonger_Settings",
  	defaults = Gossipmonger.DefaultSettings,
  	OnShow = UpdateLabel
  });

  if (oldver and oldver ~= THIS_VERSION) then
    Gossipmonger_Settings.Version = THIS_VERSION
    for k,v in pairs(Gossipmonger.DefaultSettings) do
    -- Add options new to this version, set to default:
      if (Gossipmonger_Settings[k] == nil) then
        Gossipmonger_Settings[k] = v;
      end
    end
  end
  
  UpdateStandardFilters()

  Init = nil;
end

local function skipModifierKeyHeld()
  local key = Gossipmonger_Settings.greetingModifierKey
  if (key == 0) then
    return IsAltKeyDown()
  elseif (key == 1) then
    return IsControlKeyDown()
  elseif (key == 2) then
    return IsShiftKeyDown()
  end
  -- If key is set to None (3), it always returns nil.
end


local function buildOptionsTable(tab, isQuest, ...)
  local list = tab.list and wipe(tab.list) or {}
  wipe(tab)
  local key2 = isQuest and "level" or "type"
  local count = 0
  for i = 1, select("#", ...), (isQuest and 3 or 2) do
    count = count + 1
    local a, b, c = select(i, ...)
    tab["title"..count], list[#list+1] = a, a
    tab[key2..count], list[#list+1] = b, b
    if (isQuest) then
      c = not not c
      tab["isLowLevel"..count], list[#list+1] = c, c
    end
  end
  tab.list = list
end

local AvailableQuests, ActiveQuests, GossipOptions
local FILTERS = {}

local function findFilter(arg)
  if (type(arg) == "function") then
    for i,filter in ipairs(FILTERS) do
      if (filter.func == arg) then  return filter, i;  end
    end
  elseif (type(arg) == "string") then
    for i,filter in ipairs(FILTERS) do
      if (filter.name == arg) then  return filter, i;  end
    end
  end
end

local OptionClicked_type, OptionClicked_id, OptionClicked_expect_numOptions, OptionClicked_expect_title
local function OptionClickedFollowup(NPCName, text, numAvailable, available, numActive, active, numGossip, gossip)
  if (numAvailable + numActive + numGossip == OptionClicked_expect_numOptions) then
    local tab
    if (OptionClicked_type == 1) then
      tab = available
    elseif (OptionClicked_type == 2) then
      tab = active
    elseif (OptionClicked_type == 3) then
      tab = gossip
    else
      return;
    end
    if (tab["title"..OptionClicked_id] == OptionClicked_expect_title) then
      return OptionClicked_type, OptionClicked_id
    end
  end
end


-- GLOBAL FUNCTIONS
-------------------------

function Gossipmonger.SetFilter(func, name, desc, priority, enabled)
-- Priority: Larger numbers go first. The default filter (which mimics standard WoW behavior) is set at priority 0.
-- Use nil for enabled to default to disabled if the filter didn't already exist or stay at whatever it is currently
-- set to if it does exist. Use nil for name, desc, and/or priority to use the current value(s). If the filter didn't
-- already exist, a default name (e.g. "Unnamed Filter"), a blank description, and a priority of 1 are used.
  assert(type(func) == "function", "Usage: Gossipmonger.SetFilter(func, name, desc, priority, enabled); func must be a function.")
  local tab, tabpos = findFilter(func)  -- See if filter is already used.

  local addtab
  if (tab) then
    name, desc = type(name) == "string" and name or tab.name, type(desc) == "string" and desc or tab.desc
    priority = type(priority) == "number" and priority or tab.priority
    if (enabled == nil) then  enabled = tab.enabled;  else  enabled = not not enabled;  end
    if (priority ~= tab.priority) then
      tremove(FILTERS, tabpos)
      addtab = true
    end
  else
    assert(priority == nil or type(priority) == "number", "Usage: Gossipmonger.SetFilter(func, name, desc, priority, enabled); priority must be nil or a number.")
    name = type(name) == "string" and name or L.FILTER_UNNAMED
    desc = type(desc) == "string" and desc or ""
    priority = priority or 1
    enabled = not not enabled
    tab, addtab = {}, true
  end

  tab.func, tab.name, tab.desc, tab.priority, tab.enabled = func, name, desc, priority, enabled

  if (addtab) then
    -- Find a suitable spot for a filter of this priority:
    for i,filter in ipairs(FILTERS) do
      if (filter.priority <= priority) then
        tinsert(FILTERS, i, tab)
        return;
      end
    end
    tinsert(FILTERS, tab)
  end
end

function Gossipmonger.GetFilterInfo(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then  return tab.func, tab.name, tab.desc, tab.priority, tab.enabled;  end
end

function Gossipmonger.EnableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = true
    return true
  end
end

function Gossipmonger.DisableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = false
    return true
  end
end

function Gossipmonger.OnLoad()
  tinsert(UISpecialFrames, "Gossipmonger_MirrorFrame");
  UIPanelWindows.Gossipmonger_MirrorFrame = { area = "right", pushable = 8, whileDead = 1, yoffset = -10 };
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Init);
  Gossipmonger_MirrorFrame:RegisterEvent("VARIABLES_LOADED");
  Gossipmonger.OnLoad = nil
end

function Gossipmonger.OnEvent(self, event)
  if (event == "BANKFRAME_CLOSED" or event == "BATTLEFIELDS_CLOSED" or event == "MERCHANT_CLOSED" or
      event == "TAXIMAP_CLOSED" or event == "TRAINER_CLOSED" or event == "QUEST_FINISHED" or
      event == "GOSSIP_CLOSED") then
    if (IgnoreQuestFinished and GetTitleText() == IgnoreQuestFinished and event == "QUEST_FINISHED") then
      IgnoreQuestFinished = nil
      return;
    end
    BreakMirror()
  elseif (CurrentNPC == UnitName("npc")) then
    ShowMirror()
  else
    BreakMirror()
  end
end

function Gossipmonger.GoodbyeClick()
  if (MerchantFrame:IsShown()) then
    HideUIPanel(MerchantFrame)
  elseif (BankFrame:IsShown()) then
    HideUIPanel(BankFrame)
  elseif (BattlefieldFrame:IsShown()) then
    HideUIPanel(BattlefieldFrame)
  elseif (ArenaFrame:IsShown()) then
    HideUIPanel(ArenaFrame)
  elseif (TaxiFrame:IsShown()) then
    HideUIPanel(TaxiFrame)
  elseif (ClassTrainerFrame and ClassTrainerFrame:IsShown()) then
    HideUIPanel(ClassTrainerFrame)
  elseif (GossipFrame:IsShown()) then
    HideUIPanel(GossipFrame)
  elseif (QuestFrame:IsShown()) then
    HideUIPanel(QuestFrame)
  end

  if (CurrentNPC_gossip == "binder") then
    local dialog = StaticPopup_FindVisible("CONFIRM_BINDER");
    if (dialog) then
      StaticPopup_OnClick(dialog, 2)   -- Call as if user pressed cancel button
    end
  end
end

function Gossipmonger.OptionClick(self)
  if (not SafeMirrorOptionsClick or self.selected) then  return;  end
  if (QuestFrame:IsShown()) then
    local optionType, typeid = self.type
    if (optionType == "Available") then
      typeid = 1
    elseif (optionType == "Active") then
      typeid = 2
    elseif (optionType == "Gossip") then
      typeid = 3
    else
      return;
    end

    OptionClicked_type, OptionClicked_id = typeid, self:GetID()
    OptionClicked_expect_numOptions = Gossipmonger_MirrorFrame.numOptions
    OptionClicked_expect_title = self.title
    GossipNPC = UnitName("npc")
    FollowupFunction = OptionClickedFollowup
    FollowupFunction_filtername = "<QUESTJUMPER>"

    SafeMirrorOptionsClick = nil
    DeclineQuest();
    PlaySound("igQuestCancel");
    --HideUIPanel(Gossipmonger_MirrorFrame)  -- Hide the mirror (but don't "break" it)
  end
end

--[[ DEBUG: Force the "Gossip Bubble" (AKA Mirror Frame) to show:
function Gossipmonger.ShowBubble(delay, name, text, selected, gossipType)
  name = name or UnitName("npc")
  text = text and text:gsub("||n","|n") or GetGossipText()
  showGossipMirror(delay, name, text, selected, gossipType)
end
--]]

Gossipmonger.PrepGossipOptionDisplay = prepGossipOptionDisplay

Gossipmonger.ShowMirror = ShowMirror


-- HOOK
-------------

local orig_GossipFrame_OnEvent = GossipFrame_OnEvent

local function filtererror(filter, msg, ...)
  orig_GossipFrame_OnEvent(...)
  local name = filter and '"'..filter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
  FollowupFunction = nil
  error("Invalid return from filter ("..name.."): "..msg, 2)
end

local function ShowGossipFrame(self)
  -- Show default gossip frame (see GossipFrame_OnEvent in FrameXML\GossipFrame.lua):
  if ( not GossipFrame:IsShown() ) then
    ShowUIPanel(self);
    if ( not self:IsShown() ) then
      CloseGossip();
      return;
    end
  end
  GossipFrameUpdate();

  return true
end

GossipFrame_OnEvent = function(self, event, ...)

  if (event == "GOSSIP_SHOW") then
    local name = UnitName("npc")
    local text = GetGossipText()
    local numAvailable, numActive, numGossip = GetNumGossipAvailableQuests(), GetNumGossipActiveQuests(), GetNumGossipOptions()

    local validFollowup, showMirrorIfNoFollowupResult
    if (GossipNPC == name or BlockNextFilter == name) then
    -- Dialog continued from auto-selected gossip option or player-canceled quest display:
      if (FollowupFunction and not BlockNextFilter) then
        validFollowup = FollowupFunction
        showMirrorIfNoFollowupResult = ShowMirrorOnGossip
      else
        if (not ShowGossipFrame(self)) then
          BreakMirror()
        elseif (BlockNextFilter) then
        -- This is the result of the player declining a quest or canceling the turn-in of a quest:
          if (numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
          -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
          -- the Gossipmonger "quest hanger" feature:
            showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
          else
            Gossipmonger_MirrorFrame.selected = nil
            Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
            Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
            Gossipmonger_MirrorFrame.NPCName = name
          end
        elseif (ShowMirrorOnGossip) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        end
        GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
        return;
      end
    end
    GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil

    -- New dialog or using a followup function:
    local skipdefault, keyheld = Gossipmonger_Settings.skipIsDefault, skipModifierKeyHeld()
    if ( (validFollowup or (not keyheld and skipdefault) or (keyheld and not skipdefault)) and
         (numAvailable > 0 or numActive > 0 or numGossip > 0) ) then
      AvailableQuests, ActiveQuests, GossipOptions = AvailableQuests or {}, ActiveQuests or {}, GossipOptions or {}
      buildOptionsTable(AvailableQuests, true, GetGossipAvailableQuests())
      buildOptionsTable(ActiveQuests, true, GetGossipActiveQuests())
      buildOptionsTable(GossipOptions, nil, GetGossipOptions())
      local seltype, seloption, followup, override
      local thefilter

      if (validFollowup) then
        seltype, seloption, followup, override =
          validFollowup(name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions, true)
      else
        for i,filter in ipairs(FILTERS) do
          if (filter.enabled) then
            seltype, seloption, followup, override =
              filter.func(name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions)
            if (seltype and (seloption or seltype == 0)) then
              thefilter = filter
              break
            end
          end
        end
      end

      if (seltype and seloption and seltype ~= 0 and seloption ~= 0) then  -- Filter with results found:
        -- If seltype is 0, this section is skipped, causing the default gossip frame to be shown.
        if (type(seloption) ~= "number") then
          filtererror(thefilter, "Second return (option) must be a number.", self, event, ...)
          return;
        end
        local numOptions, SelectOption, gossipType
        if (seltype == 1) then
          numOptions, SelectOption = numAvailable, SelectGossipAvailableQuest
        elseif (seltype == 2) then
          numOptions, SelectOption = numActive, SelectGossipActiveQuest
        elseif (seltype == 3) then
          numOptions, SelectOption = numGossip, SelectGossipOption
          gossipType = GossipOptions["type"..seloption]
        else
          filtererror(thefilter, "First return (type) must be an integer between 0 and 3.", self, event, ...)
          return;
        end
        if (seloption < 1 or seloption > numOptions) then
          filtererror(thefilter, "Option of that type ("..seltype..") and index ("..seloption..") not found.", self, event, ...)
          return;
        end

        local inform_SCT, inform_log = Gossipmonger_Settings.informOnSkip_SCT and SCT, Gossipmonger_Settings.informOnSkip_log
        if (override ~= false and (inform_SCT or inform_log)) then
          local named, inform = L.SKIPPINGTEXT_NAME:format(name)
          if (not gossipType or Gossipmonger_Settings.alwaysInformWithName or strlen(name) <= 20) then
            inform = named
          else
            local titled = L.SKIPPINGTEXT_TYPE:format(gossipType)
            inform = strlen(named) <= strlen(titled) and named or titled
          end
          if (inform_SCT) then
            SCTColors = SCTColors or { r = R, g = G, b = B }
            SCT:DisplayMessage(ChatIcon..inform, SCTColors)
          end
          if (inform_log) then
            chatprint(ChatIcon.." "..CYAN..inform)
          end
        end

        -- If override is true, the mirror is forced to show. If it is false (not nil), it is forced not to.
        if (override == false) then
          BreakMirror()
        elseif (override or Gossipmonger_Settings.showGossipMirrorOnSkip) then
          local index = seloption
          if (seltype > 1) then  index = index + numAvailable;  end
          if (seltype > 2) then  index = index + numActive;  end
          showGossipMirror(Gossipmonger_Settings.delayShowingMirror, name, text, index, gossipType, nil, (seltype ~= 3), numAvailable + numActive)
          ShowMirrorOnGossip = true
        end

        if (gossipType == "gossip") then
          GossipNPC = name
          if (type(followup) == "function") then
            FollowupFunction = followup
            FollowupFunction_filtername = thefilter.name
          end
        end
        SelectOption(seloption)
        return;

      elseif (showMirrorIfNoFollowupResult) then
        if (ShowGossipFrame(self)) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        else
          BreakMirror()
        end
        return;
      end
    end

    local shown = ShowGossipFrame(self)

    if (Gossipmonger_MirrorFrame:IsShown()) then
    -- Fix the odd case where the mirror is sometimes left up after it was shown without a delay and the
    -- player ran away from the NPC before the selected option's result appears.
      BreakMirror()
    end

    if (shown and numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
    -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
    -- the Gossipmonger "quest hanger" feature:
      showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
    else
      Gossipmonger_MirrorFrame.selected = nil
      Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
      Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
      Gossipmonger_MirrorFrame.NPCName = name
    end

  else
    GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
    orig_GossipFrame_OnEvent(self, event, ...)
  end
end

GossipFrame:SetScript("OnEvent", GossipFrame_OnEvent) -- Need to do this as of WoW 3.1 due to XML's function=<x> usage.

local orig_binderOnHide = StaticPopupDialogs["CONFIRM_BINDER"].OnHide;
StaticPopupDialogs["CONFIRM_BINDER"].OnHide = function(...)
  BreakMirror();
  if (orig_binderOnHide) then  orig_binderOnHide(...)  end
end

-- This function prevents Gossipmonger from running its filters when options appear due to the player hitting
-- the "Cancel" button to decline an offered quest or cancel turning in a quest.
local function QuestDeclined_DoNotFilter()
  BlockNextFilter = UnitName("npc")
end

hooksecurefunc("QuestGoodbyeButton_OnClick", QuestDeclined_DoNotFilter)
hooksecurefunc("QuestRewardCancelButton_OnClick", QuestDeclined_DoNotFilter)
hooksecurefunc("QuestDetailDeclineButton_OnClick", QuestDeclined_DoNotFilter)

-- This function prevents the mirror from disappearing if it was visible when the player hits the "Continue"
-- button (for completing a quest).
local function QuestFrameShiftFromProgressToRewards()
  if (Gossipmonger_MirrorFrame:IsShown()) then
    IgnoreQuestFinished = GetTitleText()
  else
    IgnoreQuestFinished = nil
  end
end

hooksecurefunc("QuestProgressCompleteButton_OnClick", QuestFrameShiftFromProgressToRewards)
